@@ -1,31 +1,24 @@
-use std::cell::Cell;
-
 use cairo::{self, MatrixTrait};
+use markup5ever::local_name;
 
-use crate::attributes::Attribute;
 use crate::bbox::BoundingBox;
 use crate::coord_units::CoordUnits;
 use crate::drawing_ctx::DrawingCtx;
 use crate::error::RenderingError;
-use crate::node::{NodeResult, NodeTrait, RsvgNode};
+use crate::node::{CascadedValues, NodeDraw, NodeResult, NodeTrait, RsvgNode};
 use crate::parsers::ParseValue;
 use crate::property_bag::PropertyBag;
 
 coord_units!(ClipPathUnits, CoordUnits::UserSpaceOnUse);
 
+#[derive(Default)]
 pub struct NodeClipPath {
-    units: Cell<ClipPathUnits>,
+    units: ClipPathUnits,
 }
 
 impl NodeClipPath {
-    pub fn new() -> NodeClipPath {
-        NodeClipPath {
-            units: Cell::new(ClipPathUnits::default()),
-        }
-    }
-
     pub fn get_units(&self) -> ClipPathUnits {
-        self.units.get()
+        self.units
     }
 
     pub fn to_cairo_context(
@@ -34,20 +27,18 @@ impl NodeClipPath {
         draw_ctx: &mut DrawingCtx,
         bbox: &BoundingBox,
     ) -> Result<(), RenderingError> {
-        let clip_units = self.units.get();
-
-        if clip_units == ClipPathUnits(CoordUnits::ObjectBoundingBox) && bbox.rect.is_none() {
+        if self.units == ClipPathUnits(CoordUnits::ObjectBoundingBox) && bbox.rect.is_none() {
             // The node being clipped is empty / doesn't have a
             // bounding box, so there's nothing to clip!
             return Ok(());
         }
 
-        let cascaded = node.get_cascaded_values();
+        let cascaded = CascadedValues::new_from_node(node);
 
         draw_ctx.with_saved_matrix(&mut |dc| {
             let cr = dc.get_cairo_context();
 
-            if clip_units == ClipPathUnits(CoordUnits::ObjectBoundingBox) {
+            if self.units == ClipPathUnits(CoordUnits::ObjectBoundingBox) {
                 let bbox_rect = bbox.rect.as_ref().unwrap();
 
                 cr.transform(cairo::Matrix::new(
@@ -70,10 +61,10 @@ impl NodeClipPath {
 }
 
 impl NodeTrait for NodeClipPath {
-    fn set_atts(&self, _: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::ClipPathUnits => self.units.set(attr.parse(value)?),
+                local_name!("clipPathUnits") => self.units = attr.parse(value)?,
                 _ => (),
             }
         }
