@@ -13,16 +13,20 @@ use crate::coord_units::CoordUnits;
 use crate::dpi::Dpi;
 use crate::error::RenderingError;
 use crate::filters;
-use crate::gradient::NodeGradient;
+use crate::gradient::{NodeLinearGradient, NodeRadialGradient};
 use crate::length::Dasharray;
 use crate::mask::NodeMask;
-use crate::node::{CascadedValues, NodeType, RsvgNode};
+use crate::node::{CascadedValues, NodeDraw, NodeType, RsvgNode};
 use crate::paint_server::{PaintServer, PaintSource};
 use crate::pattern::NodePattern;
 use crate::properties::ComputedValues;
 use crate::property_defs::{
-    ClipRule, FillRule, ShapeRendering, StrokeDasharray,
-    StrokeLinecap, StrokeLinejoin,
+    ClipRule,
+    FillRule,
+    ShapeRendering,
+    StrokeDasharray,
+    StrokeLinecap,
+    StrokeLinejoin,
 };
 use crate::rect::RectangleExt;
 use crate::surface_utils::shared_surface::SharedImageSurface;
@@ -221,10 +225,16 @@ impl DrawingCtx {
     }
 
     fn size_for_temporary_surface(&self) -> (i32, i32) {
+        let (viewport_width, viewport_height) = (self.rect.width, self.rect.height);
+
+        let (scaled_width, scaled_height) = self
+            .initial_affine_with_offset()
+            .transform_distance(viewport_width, viewport_height);
+
         // We need a size in whole pixels, so use ceil() to ensure the whole viewport fits
         // into the temporary surface.
-        let width = self.rect.width.ceil() as i32;
-        let height = self.rect.height.ceil() as i32;
+        let width = scaled_width.ceil() as i32;
+        let height = scaled_height.ceil() as i32;
 
         (width, height)
     }
@@ -347,8 +357,7 @@ impl DrawingCtx {
         {
             let clip_node = clip_node.get().clone();
 
-            let ClipPathUnits(units) =
-                clip_node.with_impl(|clip_path: &NodeClipPath| clip_path.get_units());
+            let ClipPathUnits(units) = clip_node.borrow().get_impl::<NodeClipPath>().get_units();
 
             if units == CoordUnits::UserSpaceOnUse {
                 (Some(clip_node), None)
@@ -362,12 +371,12 @@ impl DrawingCtx {
     }
 
     fn clip_to_node(&mut self, clip_node: &Option<RsvgNode>) -> Result<(), RenderingError> {
-        if let Some(clip_node) = clip_node {
+        if let Some(node) = clip_node {
             let orig_bbox = self.bbox;
 
-            let res = clip_node.with_impl(|clip_path: &NodeClipPath| {
-                clip_path.to_cairo_context(&clip_node, self, &orig_bbox)
-            });
+            let node_data = node.borrow();
+            let clip_path = node_data.get_impl::<NodeClipPath>();
+            let res = clip_path.to_cairo_context(&node, self, &orig_bbox);
 
             // FIXME: this is an EPIC HACK to keep the clipping context from
             // accumulating bounding boxes.  We'll remove this later, when we
@@ -396,7 +405,7 @@ impl DrawingCtx {
                 let mask = values.mask.0.get();
 
                 // The `filter` property does not apply to masks.
-                let filter = if node.get_type() == NodeType::Mask {
+                let filter = if node.borrow().get_type() == NodeType::Mask {
                     None
                 } else {
                     values.filter.0.get()
@@ -483,10 +492,11 @@ impl DrawingCtx {
                             let mask_node = acquired.get();
 
                             res = res.and_then(|_| {
-                                mask_node.with_impl(|mask: &NodeMask| {
-                                    let bbox = dc.bbox;
-                                    mask.generate_cairo_mask(&mask_node, &affines, dc, &bbox)
-                                })
+                                let bbox = dc.bbox;
+                                mask_node
+                                    .borrow()
+                                    .get_impl::<NodeMask>()
+                                    .generate_cairo_mask(&mask_node, &affines, dc, &bbox)
                             });
                         } else {
                             rsvg_log!("element {} references nonexistent mask \"{}\"", node, mask);
@@ -565,7 +575,7 @@ impl DrawingCtx {
             Some(acquired) => {
                 let filter_node = acquired.get();
 
-                if !filter_node.is_in_error() {
+                if !filter_node.borrow().is_in_error() {
                     // FIXME: deal with out of memory here
                     filters::render(&filter_node, values, child_surface, self, node_bbox)
                 } else {
@@ -618,28 +628,30 @@ impl DrawingCtx {
         bbox: &BoundingBox,
         current_color: &cssparser::RGBA,
     ) -> Result<bool, RenderingError> {
-        let mut had_paint_server;
-
         match *ps {
             PaintServer::Iri {
                 ref iri,
                 ref alternate,
             } => {
-                had_paint_server = false;
+                let mut had_paint_server = false;
 
                 if let Some(acquired) = self.acquired_nodes.get_node(iri) {
                     let node = acquired.get();
 
-                    if node.get_type() == NodeType::LinearGradient
-                        || node.get_type() == NodeType::RadialGradient
-                    {
-                        had_paint_server = node.with_impl(|n: &NodeGradient| {
-                            n.resolve_fallbacks_and_set_pattern(&node, self, opacity, bbox)
-                        })?;
-                    } else if node.get_type() == NodeType::Pattern {
-                        had_paint_server = node.with_impl(|n: &NodePattern| {
-                            n.resolve_fallbacks_and_set_pattern(&node, self, opacity, bbox)
-                        })?;
+                    had_paint_server = match node.borrow().get_type() {
+                        NodeType::LinearGradient => node
+                            .borrow()
+                            .get_impl::<NodeLinearGradient>()
+                            .resolve_fallbacks_and_set_pattern(&node, self, opacity, bbox)?,
+                        NodeType::RadialGradient => node
+                            .borrow()
+                            .get_impl::<NodeRadialGradient>()
+                            .resolve_fallbacks_and_set_pattern(&node, self, opacity, bbox)?,
+                        NodeType::Pattern => node
+                            .borrow()
+                            .get_impl::<NodePattern>()
+                            .resolve_fallbacks_and_set_pattern(&node, self, opacity, bbox)?,
+                        _ => false,
                     }
                 }
 
@@ -652,19 +664,17 @@ impl DrawingCtx {
                         iri
                     );
                 }
+
+                Ok(had_paint_server)
             }
 
             PaintServer::SolidColor(color) => {
                 self.set_color(&color, opacity, current_color);
-                had_paint_server = true;
+                Ok(true)
             }
 
-            PaintServer::None => {
-                had_paint_server = false;
-            }
-        };
-
-        Ok(had_paint_server)
+            PaintServer::None => Ok(false),
+        }
     }
 
     pub fn setup_cr_for_stroke(&self, cr: &cairo::Context, values: &ComputedValues) {
@@ -837,7 +847,7 @@ impl DrawingCtx {
 
         let values = cascaded.get();
         let res = if draw && values.is_visible() {
-            node.draw(node, cascaded, self, clipping)
+            node.draw(cascaded, self, clipping)
         } else {
             Ok(())
         };
@@ -869,6 +879,7 @@ impl DrawingCtx {
     }
 }
 
+#[derive(Debug)]
 pub struct CompositingAffines {
     pub outside_temporary_surface: cairo::Matrix,
     pub initial: cairo::Matrix,
@@ -893,26 +904,25 @@ impl CompositingAffines {
             cairo::Matrix::multiply(&current, &initial_inverse)
         };
 
+        let (scale_x, scale_y) = initial.transform_distance(1.0, 1.0);
+
         let for_temporary_surface = if is_topmost_temporary_surface {
             let untransformed = cairo::Matrix::multiply(&current, &initial_inverse);
-            untransformed
+            let scale = cairo::Matrix::new(scale_x, 0.0, 0.0, scale_y, 0.0, 0.0);
+            cairo::Matrix::multiply(&untransformed, &scale)
         } else {
             current
         };
 
         let compositing = if is_topmost_temporary_surface {
-            initial
+            let mut scaled = initial;
+            scaled.scale(1.0 / scale_x, 1.0 / scale_y);
+            scaled
         } else {
             cairo::Matrix::identity()
         };
 
-        // This is the inverse of "compositing"; we do it this way
-        // instead of inverting that one to preserve accuracy.
-        let for_snapshot = if is_topmost_temporary_surface {
-            initial_inverse
-        } else {
-            cairo::Matrix::identity()
-        };
+        let for_snapshot = compositing.try_invert().unwrap();
 
         CompositingAffines {
             outside_temporary_surface,
@@ -1139,7 +1149,7 @@ impl AcquiredNodes {
         fragment
             .and_then(move |fragment| self.get_node(fragment))
             .and_then(|acquired| {
-                if acquired.get().get_type() == node_type {
+                if acquired.get().borrow().get_type() == node_type {
                     Some(acquired)
                 } else {
                     None
