@@ -114,9 +114,10 @@ impl FilterContext {
             height: 0.0,
         });
 
-        let filter = filter_node.get_impl::<NodeFilter>().unwrap();
+        let node_data = filter_node.borrow();
+        let filter = node_data.get_impl::<NodeFilter>();
 
-        let affine = match filter.filterunits.get() {
+        let affine = match filter.get_filter_units() {
             CoordUnits::UserSpaceOnUse => cr_affine,
             CoordUnits::ObjectBoundingBox => {
                 let affine = cairo::Matrix::new(
@@ -131,7 +132,7 @@ impl FilterContext {
             }
         };
 
-        let paffine = match filter.primitiveunits.get() {
+        let paffine = match filter.get_primitive_units() {
             CoordUnits::UserSpaceOnUse => cr_affine,
             CoordUnits::ObjectBoundingBox => {
                 let affine = cairo::Matrix::new(
@@ -305,10 +306,11 @@ impl FilterContext {
 
     /// Pushes the viewport size based on the value of `primitiveUnits`.
     pub fn get_view_params(&self, draw_ctx: &mut DrawingCtx) -> ViewParams {
-        let filter = self.node.get_impl::<NodeFilter>().unwrap();
+        let node_data = self.node.borrow();
+        let filter = node_data.get_impl::<NodeFilter>();
 
         // See comments in compute_effects_region() for how this works.
-        if filter.primitiveunits.get() == CoordUnits::ObjectBoundingBox {
+        if filter.get_primitive_units() == CoordUnits::ObjectBoundingBox {
             draw_ctx.push_view_box(1.0, 1.0)
         } else {
             draw_ctx.get_view_params()
