@@ -1,10 +1,8 @@
-use std::cell::{Cell, RefCell};
-
 use cairo::{self, ImageSurface, MatrixTrait};
+use markup5ever::{local_name, LocalName};
 
-use crate::attributes::Attribute;
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::NodeError;
+use crate::error::{AttributeResultExt, NodeError};
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
 use crate::parsers::{self, ParseError};
 use crate::property_bag::PropertyBag;
@@ -25,44 +23,42 @@ enum ColorChannel {
 /// The `feDisplacementMap` filter primitive.
 pub struct DisplacementMap {
     base: PrimitiveWithInput,
-    in2: RefCell<Option<Input>>,
-    scale: Cell<f64>,
-    x_channel_selector: Cell<ColorChannel>,
-    y_channel_selector: Cell<ColorChannel>,
+    in2: Option<Input>,
+    scale: f64,
+    x_channel_selector: ColorChannel,
+    y_channel_selector: ColorChannel,
 }
 
-impl DisplacementMap {
+impl Default for DisplacementMap {
     /// Constructs a new `DisplacementMap` with empty properties.
     #[inline]
-    pub fn new() -> DisplacementMap {
+    fn default() -> DisplacementMap {
         DisplacementMap {
             base: PrimitiveWithInput::new::<Self>(),
-            in2: RefCell::new(None),
-            scale: Cell::new(0.0),
-            x_channel_selector: Cell::new(ColorChannel::A),
-            y_channel_selector: Cell::new(ColorChannel::A),
+            in2: None,
+            scale: 0.0,
+            x_channel_selector: ColorChannel::A,
+            y_channel_selector: ColorChannel::A,
         }
     }
 }
 
 impl NodeTrait for DisplacementMap {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.base.set_atts(node, pbag)?;
+    impl_node_as_filter!();
+
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        self.base.set_atts(parent, pbag)?;
 
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::In2 => {
-                    self.in2.replace(Some(Input::parse(Attribute::In2, value)?));
+                local_name!("in2") => self.in2 = Some(Input::parse(attr, value)?),
+                local_name!("scale") => self.scale = parsers::number(value).attribute(attr)?,
+                local_name!("xChannelSelector") => {
+                    self.x_channel_selector = ColorChannel::parse(attr, value)?
+                }
+                local_name!("yChannelSelector") => {
+                    self.y_channel_selector = ColorChannel::parse(attr, value)?
                 }
-                Attribute::Scale => self.scale.set(
-                    parsers::number(value).map_err(|err| NodeError::attribute_error(attr, err))?,
-                ),
-                Attribute::XChannelSelector => self
-                    .x_channel_selector
-                    .set(ColorChannel::parse(attr, value)?),
-                Attribute::YChannelSelector => self
-                    .y_channel_selector
-                    .set(ColorChannel::parse(attr, value)?),
                 _ => (),
             }
         }
@@ -79,7 +75,7 @@ impl Filter for DisplacementMap {
         draw_ctx: &mut DrawingCtx,
     ) -> Result<FilterResult, FilterError> {
         let input = self.base.get_input(ctx, draw_ctx)?;
-        let displacement_input = ctx.get_input(draw_ctx, self.in2.borrow().as_ref())?;
+        let displacement_input = ctx.get_input(draw_ctx, self.in2.as_ref())?;
         let bounds = self
             .base
             .get_bounds(ctx)
@@ -90,11 +86,7 @@ impl Filter for DisplacementMap {
         // Displacement map's values need to be non-premultiplied.
         let displacement_surface = displacement_input.surface().unpremultiply(bounds)?;
 
-        let scale = self.scale.get();
-        let (sx, sy) = ctx.paffine().transform_distance(scale, scale);
-
-        let x_channel = self.x_channel_selector.get();
-        let y_channel = self.y_channel_selector.get();
+        let (sx, sy) = ctx.paffine().transform_distance(self.scale, self.scale);
 
         let output_surface = ImageSurface::create(
             cairo::Format::ARgb32,
@@ -115,8 +107,8 @@ impl Filter for DisplacementMap {
 
                 let process = |x| f64::from(x) / 255.0 - 0.5;
 
-                let dx = process(get_value(x_channel));
-                let dy = process(get_value(y_channel));
+                let dx = process(get_value(self.x_channel_selector));
+                let dy = process(get_value(self.y_channel_selector));
 
                 let x = f64::from(x);
                 let y = f64::from(y);
@@ -135,7 +127,7 @@ impl Filter for DisplacementMap {
         }
 
         Ok(FilterResult {
-            name: self.base.result.borrow().clone(),
+            name: self.base.result.clone(),
             output: FilterOutput {
                 surface: SharedImageSurface::new(output_surface, input.surface().surface_type())?,
                 bounds,
@@ -152,7 +144,7 @@ impl Filter for DisplacementMap {
 }
 
 impl ColorChannel {
-    fn parse(attr: Attribute, s: &str) -> Result<Self, NodeError> {
+    fn parse(attr: LocalName, s: &str) -> Result<Self, NodeError> {
         match s {
             "R" => Ok(ColorChannel::R),
             "G" => Ok(ColorChannel::G),
