@@ -1,13 +1,12 @@
-use std::cell::Cell;
 use std::cmp::min;
 use std::f64;
 
 use cairo::MatrixTrait;
+use markup5ever::local_name;
 use nalgebra::{DMatrix, Dynamic, VecStorage};
 
-use crate::attributes::Attribute;
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::NodeError;
+use crate::error::{AttributeResultExt, NodeError};
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
 use crate::parsers;
 use crate::property_bag::PropertyBag;
@@ -28,37 +27,39 @@ const MAXIMUM_KERNEL_SIZE: usize = 500;
 /// The `feGaussianBlur` filter primitive.
 pub struct GaussianBlur {
     base: PrimitiveWithInput,
-    std_deviation: Cell<(f64, f64)>,
+    std_deviation: (f64, f64),
 }
 
-impl GaussianBlur {
+impl Default for GaussianBlur {
     /// Constructs a new `GaussianBlur` with empty properties.
     #[inline]
-    pub fn new() -> GaussianBlur {
+    fn default() -> GaussianBlur {
         GaussianBlur {
             base: PrimitiveWithInput::new::<Self>(),
-            std_deviation: Cell::new((0.0, 0.0)),
+            std_deviation: (0.0, 0.0),
         }
     }
 }
 
 impl NodeTrait for GaussianBlur {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.base.set_atts(node, pbag)?;
+    impl_node_as_filter!();
+
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        self.base.set_atts(parent, pbag)?;
 
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::StdDeviation => self.std_deviation.set(
-                    parsers::number_optional_number(value)
-                        .map_err(|err| NodeError::attribute_error(attr, err))
+                local_name!("stdDeviation") => {
+                    self.std_deviation = parsers::number_optional_number(value)
+                        .attribute(attr.clone())
                         .and_then(|(x, y)| {
                             if x >= 0.0 && y >= 0.0 {
                                 Ok((x, y))
                             } else {
                                 Err(NodeError::value_error(attr, "values can't be negative"))
                             }
-                        })?,
-                ),
+                        })?
+                }
                 _ => (),
             }
         }
@@ -210,7 +211,7 @@ impl Filter for GaussianBlur {
             .add_input(&input)
             .into_irect(draw_ctx);
 
-        let (std_x, std_y) = self.std_deviation.get();
+        let (std_x, std_y) = self.std_deviation;
         let (std_x, std_y) = ctx.paffine().transform_distance(std_x, std_y);
 
         // The deviation can become negative here due to the transform.
@@ -246,7 +247,7 @@ impl Filter for GaussianBlur {
         };
 
         Ok(FilterResult {
-            name: self.base.result.borrow().clone(),
+            name: self.base.result.clone(),
             output: FilterOutput {
                 surface: output_surface,
                 bounds,
