@@ -1,10 +1,8 @@
-use std::cell::{Cell, RefCell};
-
 use cairo::{self, ImageSurface, MatrixTrait, PatternTrait, Rectangle};
+use markup5ever::local_name;
 
 use crate::allowed_url::{Fragment, Href};
 use crate::aspect_ratio::AspectRatio;
-use crate::attributes::Attribute;
 use crate::drawing_ctx::DrawingCtx;
 use crate::error::{NodeError, RenderingError};
 use crate::float_eq_cairo::ApproxEqCairo;
@@ -21,21 +19,23 @@ use super::{Filter, FilterError, Primitive};
 /// The `feImage` filter primitive.
 pub struct Image {
     base: Primitive,
-    aspect: Cell<AspectRatio>,
-    href: RefCell<Option<Href>>,
+    aspect: AspectRatio,
+    href: Option<Href>,
+}
+
+impl Default for Image {
+    /// Constructs a new `Image` with empty properties.
+    #[inline]
+    fn default() -> Image {
+        Image {
+            base: Primitive::new::<Self>(),
+            aspect: AspectRatio::default(),
+            href: None,
+        }
+    }
 }
 
 impl Image {
-    /// Constructs a new `Image` with empty properties.
-    #[inline]
-    pub fn new() -> Image {
-        Image {
-            base: Primitive::new::<Self>(),
-            aspect: Cell::new(AspectRatio::default()),
-            href: RefCell::new(None),
-        }
-    }
-
     /// Renders the filter if the source is an existing node.
     fn render_node(
         &self,
@@ -126,8 +126,7 @@ impl Image {
         )?;
 
         // TODO: this goes through a f64->i32->f64 conversion.
-        let aspect = self.aspect.get();
-        let (x, y, w, h) = aspect.compute(
+        let (x, y, w, h) = self.aspect.compute(
             &ViewBox::new(
                 0.0,
                 0.0,
@@ -174,20 +173,22 @@ impl Image {
 }
 
 impl NodeTrait for Image {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.base.set_atts(node, pbag)?;
+    impl_node_as_filter!();
+
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        self.base.set_atts(parent, pbag)?;
 
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::PreserveAspectRatio => self.aspect.set(attr.parse(value)?),
+                local_name!("preserveAspectRatio") => self.aspect = attr.parse(value)?,
 
                 // "path" is used by some older Adobe Illustrator versions
-                Attribute::XlinkHref | Attribute::Path => {
+                local_name!("xlink:href") | local_name!("path") => {
                     let href = Href::parse(value).map_err(|_| {
                         NodeError::parse_error(attr, ParseError::new("could not parse href"))
                     })?;
 
-                    *self.href.borrow_mut() = Some(href);
+                    self.href = Some(href);
                 }
 
                 _ => (),
@@ -208,11 +209,8 @@ impl Filter for Image {
         let bounds_builder = self.base.get_bounds(ctx);
         let bounds = bounds_builder.into_irect(draw_ctx);
 
-        let href_borrow = self.href.borrow();
-        let href_opt = href_borrow.as_ref();
-
-        if let Some(href) = href_opt {
-            let output_surface = match *href {
+        if let Some(href) = self.href.as_ref() {
+            let output_surface = match href {
                 Href::PlainUrl(_) => {
                     let unclipped_bounds = bounds_builder.into_irect_without_clipping(draw_ctx);
                     self.render_external_image(ctx, draw_ctx, &bounds, &unclipped_bounds, href)?
@@ -221,7 +219,7 @@ impl Filter for Image {
             };
 
             Ok(FilterResult {
-                name: self.base.result.borrow().clone(),
+                name: self.base.result.clone(),
                 output: FilterOutput {
                     surface: SharedImageSurface::new(output_surface, SurfaceType::SRgb)?,
                     bounds,
