@@ -1,23 +1,20 @@
-use std::cell::{Cell, RefCell};
 use std::ops::Deref;
 use std::time::Instant;
 
 use cairo::{self, MatrixTrait};
-use owning_ref::RcRef;
+use markup5ever::local_name;
 
-use crate::attributes::Attribute;
 use crate::bbox::BoundingBox;
 use crate::coord_units::CoordUnits;
 use crate::drawing_ctx::DrawingCtx;
 use crate::error::{RenderingError, ValueErrorKind};
 use crate::length::{LengthHorizontal, LengthUnit, LengthVertical};
-use crate::node::{NodeData, NodeResult, NodeTrait, NodeType, RsvgNode};
+use crate::node::{CascadedValues, NodeResult, NodeTrait, NodeType, RsvgNode};
 use crate::parsers::{ParseError, ParseValue};
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
 use crate::property_defs::ColorInterpolationFilters;
 use crate::surface_utils::shared_surface::{SharedImageSurface, SurfaceType};
-use crate::tree_utils::{Node, NodeRef};
 
 mod bounds;
 use self::bounds::BoundsBuilder;
@@ -34,24 +31,8 @@ use self::input::Input;
 pub mod node;
 use self::node::NodeFilter;
 
-pub mod blend;
-pub mod color_matrix;
-pub mod component_transfer;
-pub mod composite;
-pub mod convolve_matrix;
-pub mod displacement_map;
-pub mod flood;
-pub mod gaussian_blur;
-pub mod image;
-pub mod light;
-pub mod merge;
-pub mod morphology;
-pub mod offset;
-pub mod tile;
-pub mod turbulence;
-
 /// A filter primitive interface.
-trait Filter: NodeTrait {
+pub trait Filter: NodeTrait {
     /// Renders this filter primitive.
     ///
     /// If this filter primitive can't be rendered for whatever reason (for instance, a required
@@ -71,19 +52,43 @@ trait Filter: NodeTrait {
     fn is_affected_by_color_interpolation_filters(&self) -> bool;
 }
 
+macro_rules! impl_node_as_filter {
+    () => (
+        fn as_filter(&self) -> Option<&Filter> {
+            Some(self)
+        }
+    )
+}
+
+pub mod blend;
+pub mod color_matrix;
+pub mod component_transfer;
+pub mod composite;
+pub mod convolve_matrix;
+pub mod displacement_map;
+pub mod flood;
+pub mod gaussian_blur;
+pub mod image;
+pub mod light;
+pub mod merge;
+pub mod morphology;
+pub mod offset;
+pub mod tile;
+pub mod turbulence;
+
 /// The base filter primitive node containing common properties.
 struct Primitive {
-    x: Cell<Option<LengthHorizontal>>,
-    y: Cell<Option<LengthVertical>>,
-    width: Cell<Option<LengthHorizontal>>,
-    height: Cell<Option<LengthVertical>>,
-    result: RefCell<Option<String>>,
+    x: Option<LengthHorizontal>,
+    y: Option<LengthVertical>,
+    width: Option<LengthHorizontal>,
+    height: Option<LengthVertical>,
+    result: Option<String>,
 }
 
 /// The base node for filter primitives which accept input.
 struct PrimitiveWithInput {
     base: Primitive,
-    in_: RefCell<Option<Input>>,
+    in_: Option<Input>,
 }
 
 impl Primitive {
@@ -91,35 +96,28 @@ impl Primitive {
     #[inline]
     fn new<T: Filter>() -> Primitive {
         Primitive {
-            x: Cell::new(None),
-            y: Cell::new(None),
-            width: Cell::new(None),
-            height: Cell::new(None),
-            result: RefCell::new(None),
+            x: None,
+            y: None,
+            width: None,
+            height: None,
+            result: None,
         }
     }
 
     /// Returns the `BoundsBuilder` for bounds computation.
     #[inline]
     fn get_bounds<'a>(&self, ctx: &'a FilterContext) -> BoundsBuilder<'a> {
-        BoundsBuilder::new(
-            ctx,
-            self.x.get(),
-            self.y.get(),
-            self.width.get(),
-            self.height.get(),
-        )
+        BoundsBuilder::new(ctx, self.x, self.y, self.width, self.height)
     }
 }
 
 impl NodeTrait for Primitive {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         // With ObjectBoundingBox, only fractions and percents are allowed.
-        let primitiveunits = node
-            .parent()
+        let primitiveunits = parent
             .and_then(|parent| {
-                if parent.get_type() == NodeType::Filter {
-                    Some(parent.with_impl(|f: &NodeFilter| f.primitiveunits.get()))
+                if parent.borrow().get_type() == NodeType::Filter {
+                    Some(parent.borrow().get_impl::<NodeFilter>().get_primitive_units())
                 } else {
                     None
                 }
@@ -164,19 +162,27 @@ impl NodeTrait for Primitive {
 
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::X => self.x.set(Some(
-                    attr.parse_and_validate(value, check_units_horizontal)?,
-                )),
-                Attribute::Y => self
-                    .y
-                    .set(Some(attr.parse_and_validate(value, check_units_vertical)?)),
-                Attribute::Width => self.width.set(Some(
-                    attr.parse_and_validate(value, check_units_horizontal_and_ensure_nonnegative)?,
-                )),
-                Attribute::Height => self.height.set(Some(
-                    attr.parse_and_validate(value, check_units_vertical_and_ensure_nonnegative)?,
-                )),
-                Attribute::Result => *self.result.borrow_mut() = Some(value.to_string()),
+                local_name!("x") => {
+                    self.x = Some(attr.parse_and_validate(value, check_units_horizontal)?)
+                }
+                local_name!("y") => {
+                    self.y = Some(attr.parse_and_validate(value, check_units_vertical)?)
+                }
+                local_name!("width") => {
+                    self.width =
+                        Some(attr.parse_and_validate(
+                            value,
+                            check_units_horizontal_and_ensure_nonnegative,
+                        )?)
+                }
+                local_name!("height") => {
+                    self.height =
+                        Some(attr.parse_and_validate(
+                            value,
+                            check_units_vertical_and_ensure_nonnegative,
+                        )?)
+                }
+                local_name!("result") => self.result = Some(value.to_string()),
                 _ => (),
             }
         }
@@ -191,7 +197,7 @@ impl PrimitiveWithInput {
     fn new<T: Filter>() -> PrimitiveWithInput {
         PrimitiveWithInput {
             base: Primitive::new::<T>(),
-            in_: RefCell::new(None),
+            in_: None,
         }
     }
 
@@ -202,17 +208,17 @@ impl PrimitiveWithInput {
         ctx: &FilterContext,
         draw_ctx: &mut DrawingCtx,
     ) -> Result<FilterInput, FilterError> {
-        ctx.get_input(draw_ctx, self.in_.borrow().as_ref())
+        ctx.get_input(draw_ctx, self.in_.as_ref())
     }
 }
 
 impl NodeTrait for PrimitiveWithInput {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.base.set_atts(node, pbag)?;
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        self.base.set_atts(parent, pbag)?;
 
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::In => drop(self.in_.replace(Some(Input::parse(Attribute::In, value)?))),
+                local_name!("in") => drop(self.in_ = Some(Input::parse(attr, value)?)),
                 _ => (),
             }
         }
@@ -257,8 +263,8 @@ pub fn render(
     node_bbox: BoundingBox,
 ) -> Result<cairo::ImageSurface, RenderingError> {
     let filter_node = &*filter_node;
-    assert_eq!(filter_node.get_type(), NodeType::Filter);
-    assert!(!filter_node.is_in_error());
+    assert_eq!(filter_node.borrow().get_type(), NodeType::Filter);
+    assert!(!filter_node.borrow().is_in_error());
 
     // The source surface has multiple references. We need to copy it to a new surface to have a
     // unique reference to be able to safely access the pixel data.
@@ -282,7 +288,7 @@ pub fn render(
         .children()
         // Skip nodes in error.
         .filter(|c| {
-            let in_error = c.is_in_error();
+            let in_error = c.borrow().is_in_error();
 
             if in_error {
                 rsvg_log!("(ignoring filter primitive {} because it is in error)", c);
@@ -290,78 +296,30 @@ pub fn render(
 
             !in_error
         })
+        // Keep only filter primitives (those that implement the Filter trait)
+        .filter(|c| c.borrow().get_node_trait().as_filter().is_some())
         // Check if the node wants linear RGB.
         .map(|c| {
             let linear_rgb = {
-                let cascaded = c.get_cascaded_values();
+                let cascaded = CascadedValues::new_from_node(&c);
                 let values = cascaded.get();
 
                 values.color_interpolation_filters == ColorInterpolationFilters::LinearRgb
             };
 
             (c, linear_rgb)
-        })
-        // Keep only filter primitives (those that implement the Filter trait)
-        .filter_map(|(c, linear_rgb)| {
-            let rr = RcRef::new(c.0)
-                .try_map(|c: &Node<NodeData>| {
-                    // Go through the filter primitives and see if the node is one of them.
-                    #[inline]
-                    fn as_filter<T: Filter>(x: &T) -> &Filter {
-                        x
-                    }
-
-                    // Unfortunately it's not possible to downcast to a trait object.  So, we
-                    // try downcasting to each filter type individually, and use
-                    // owning_ref::RcRef to reduce the children to a list of
-                    // OwningRef<Rc<Node<NodeData>>, dyn Filter> -- which is essentially the
-                    // NodeRef, and something inside the NodeRef that implements the Filter
-                    // trait.
-                    macro_rules! try_downcasting_to_filter {
-                        ($c:expr; $($t:ty),+$(,)*) => ({
-                            let mut filter = None;
-                            $(
-                                filter = filter.or_else(|| $c.borrow().get_impl::<$t>().map(as_filter));
-                            )+
-                            filter
-                        })
-                    }
-
-                    let filter = try_downcasting_to_filter!(c;
-                        blend::Blend,
-                        color_matrix::ColorMatrix,
-                        component_transfer::ComponentTransfer,
-                        composite::Composite,
-                        convolve_matrix::ConvolveMatrix,
-                        displacement_map::DisplacementMap,
-                        flood::Flood,
-                        gaussian_blur::GaussianBlur,
-                        image::Image,
-                        light::lighting::Lighting,
-                        merge::Merge,
-                        morphology::Morphology,
-                        offset::Offset,
-                        tile::Tile,
-                        turbulence::Turbulence,
-                    );
-                    filter.ok_or(())
-                })
-                .ok();
-
-            rr.map(|rr| (rr, linear_rgb))
         });
 
-    for (rr, linear_rgb) in primitives {
-        // rr: OwningRef<Rc<Node<NodeData>>, dyn Filter>
-
-        let rr_node = NodeRef(rr.as_owner().clone());
+    for (c, linear_rgb) in primitives {
+        let node_data = c.borrow();
+        let filter = node_data.get_node_trait().as_filter().unwrap();
 
         let mut render = |filter_ctx: &mut FilterContext| {
-            if let Err(err) = rr
-                .render(&rr_node, filter_ctx, draw_ctx)
+            if let Err(err) = filter
+                .render(&c, filter_ctx, draw_ctx)
                 .and_then(|result| filter_ctx.store_result(result))
             {
-                rsvg_log!("(filter primitive {} returned an error: {})", rr_node, err);
+                rsvg_log!("(filter primitive {} returned an error: {})", c, err);
 
                 // Exit early on Cairo errors. Continue rendering otherwise.
                 if let FilterError::CairoError(status) = err {
@@ -374,7 +332,7 @@ pub fn render(
 
         let start = Instant::now();
 
-        if rr.is_affected_by_color_interpolation_filters() && linear_rgb {
+        if filter.is_affected_by_color_interpolation_filters() && linear_rgb {
             filter_ctx.with_linear_rgb(render)?;
         } else {
             render(&mut filter_ctx)?;
@@ -383,7 +341,7 @@ pub fn render(
         let elapsed = start.elapsed();
         rsvg_log!(
             "(rendered filter primitive {} in\n    {} seconds)",
-            rr_node,
+            c,
             elapsed.as_secs() as f64 + f64::from(elapsed.subsec_nanos()) / 1e9
         );
     }
