@@ -1,10 +1,8 @@
-use std::cell::Cell;
-
 use cairo::{self, ImageSurface, MatrixTrait};
+use markup5ever::local_name;
 
-use crate::attributes::Attribute;
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::NodeError;
+use crate::error::AttributeResultExt;
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
 use crate::parsers;
 use crate::property_bag::PropertyBag;
@@ -18,34 +16,32 @@ use super::{Filter, FilterError, PrimitiveWithInput};
 /// The `feOffset` filter primitive.
 pub struct Offset {
     base: PrimitiveWithInput,
-    dx: Cell<f64>,
-    dy: Cell<f64>,
+    dx: f64,
+    dy: f64,
 }
 
-impl Offset {
+impl Default for Offset {
     /// Constructs a new `Offset` with empty properties.
     #[inline]
-    pub fn new() -> Offset {
+    fn default() -> Offset {
         Offset {
             base: PrimitiveWithInput::new::<Self>(),
-            dx: Cell::new(0f64),
-            dy: Cell::new(0f64),
+            dx: 0f64,
+            dy: 0f64,
         }
     }
 }
 
 impl NodeTrait for Offset {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.base.set_atts(node, pbag)?;
+    impl_node_as_filter!();
+
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        self.base.set_atts(parent, pbag)?;
 
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::Dx => self.dx.set(
-                    parsers::number(value).map_err(|err| NodeError::attribute_error(attr, err))?,
-                ),
-                Attribute::Dy => self.dy.set(
-                    parsers::number(value).map_err(|err| NodeError::attribute_error(attr, err))?,
-                ),
+                local_name!("dx") => self.dx = parsers::number(value).attribute(attr)?,
+                local_name!("dy") => self.dy = parsers::number(value).attribute(attr)?,
                 _ => (),
             }
         }
@@ -68,9 +64,7 @@ impl Filter for Offset {
             .add_input(&input)
             .into_irect(draw_ctx);
 
-        let dx = self.dx.get();
-        let dy = self.dy.get();
-        let (ox, oy) = ctx.paffine().transform_distance(dx, dy);
+        let (ox, oy) = ctx.paffine().transform_distance(self.dx, self.dy);
 
         // output_bounds contains all pixels within bounds,
         // for which (x - ox) and (y - oy) also lie within bounds.
@@ -102,7 +96,7 @@ impl Filter for Offset {
         }
 
         Ok(FilterResult {
-            name: self.base.result.borrow().clone(),
+            name: self.base.result.clone(),
             output: FilterOutput {
                 surface: SharedImageSurface::new(output_surface, input.surface().surface_type())?,
                 bounds,
