@@ -13,10 +13,10 @@ pub struct Tile {
     base: PrimitiveWithInput,
 }
 
-impl Tile {
+impl Default for Tile {
     /// Constructs a new `Tile` with empty properties.
     #[inline]
-    pub fn new() -> Tile {
+    fn default() -> Tile {
         Tile {
             base: PrimitiveWithInput::new::<Self>(),
         }
@@ -24,8 +24,10 @@ impl Tile {
 }
 
 impl NodeTrait for Tile {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.base.set_atts(node, pbag)
+    impl_node_as_filter!();
+
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        self.base.set_atts(parent, pbag)
     }
 }
 
@@ -96,7 +98,7 @@ impl Filter for Tile {
         };
 
         Ok(FilterResult {
-            name: self.base.result.borrow().clone(),
+            name: self.base.result.clone(),
             output: FilterOutput {
                 surface: output_surface,
                 bounds,
