@@ -1,14 +1,12 @@
-use std::cell::Cell;
-
 use cairo::{self, ImageSurface, MatrixTrait};
+use markup5ever::{local_name, LocalName};
 
-use crate::attributes::Attribute;
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::NodeError;
-use crate::node::{NodeResult, NodeTrait, RsvgNode};
+use crate::error::{AttributeResultExt, NodeError};
+use crate::node::{CascadedValues, NodeResult, NodeTrait, RsvgNode};
 use crate::parsers::{self, ParseError};
-use crate::property_defs::ColorInterpolationFilters;
 use crate::property_bag::PropertyBag;
+use crate::property_defs::ColorInterpolationFilters;
 use crate::surface_utils::{
     shared_surface::{SharedImageSurface, SurfaceType},
     ImageSurfaceDataExt,
@@ -36,52 +34,54 @@ enum NoiseType {
 /// The `feTurbulence` filter primitive.
 pub struct Turbulence {
     base: Primitive,
-    base_frequency: Cell<(f64, f64)>,
-    num_octaves: Cell<i32>,
-    seed: Cell<i32>,
-    stitch_tiles: Cell<StitchTiles>,
-    type_: Cell<NoiseType>,
+    base_frequency: (f64, f64),
+    num_octaves: i32,
+    seed: i32,
+    stitch_tiles: StitchTiles,
+    type_: NoiseType,
 }
 
-impl Turbulence {
+impl Default for Turbulence {
     /// Constructs a new `Turbulence` with empty properties.
     #[inline]
-    pub fn new() -> Turbulence {
+    fn default() -> Turbulence {
         Turbulence {
             base: Primitive::new::<Self>(),
-            base_frequency: Cell::new((0.0, 0.0)),
-            num_octaves: Cell::new(1),
-            seed: Cell::new(0),
-            stitch_tiles: Cell::new(StitchTiles::NoStitch),
-            type_: Cell::new(NoiseType::Turbulence),
+            base_frequency: (0.0, 0.0),
+            num_octaves: 1,
+            seed: 0,
+            stitch_tiles: StitchTiles::NoStitch,
+            type_: NoiseType::Turbulence,
         }
     }
 }
 
 impl NodeTrait for Turbulence {
+    impl_node_as_filter!();
+
     #[inline]
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.base.set_atts(node, pbag)?;
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        self.base.set_atts(parent, pbag)?;
 
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::BaseFrequency => self.base_frequency.set(
-                    parsers::number_optional_number(value)
-                        .map_err(|err| NodeError::attribute_error(attr, err))
+                local_name!("baseFrequency") => {
+                    self.base_frequency = parsers::number_optional_number(value)
+                        .attribute(attr.clone())
                         .and_then(|(x, y)| {
                             if x >= 0.0 && y >= 0.0 {
                                 Ok((x, y))
                             } else {
                                 Err(NodeError::value_error(attr, "values can't be negative"))
                             }
-                        })?,
-                ),
-                Attribute::NumOctaves => self.num_octaves.set(
-                    parsers::integer(value).map_err(|err| NodeError::attribute_error(attr, err))?,
-                ),
+                        })?
+                }
+                local_name!("numOctaves") => {
+                    self.num_octaves = parsers::integer(value).attribute(attr)?
+                }
                 // Yes, seed needs to be parsed as a number and then truncated.
-                Attribute::Seed => self.seed.set(
-                    parsers::number(value)
+                local_name!("seed") => {
+                    self.seed = parsers::number(value)
                         .map(|x| {
                             clamp(
                                 x.trunc(),
@@ -89,10 +89,10 @@ impl NodeTrait for Turbulence {
                                 f64::from(i32::max_value()),
                             ) as i32
                         })
-                        .map_err(|err| NodeError::attribute_error(attr, err))?,
-                ),
-                Attribute::StitchTiles => self.stitch_tiles.set(StitchTiles::parse(attr, value)?),
-                Attribute::Type => self.type_.set(NoiseType::parse(attr, value)?),
+                        .attribute(attr)?
+                }
+                local_name!("stitchTiles") => self.stitch_tiles = StitchTiles::parse(attr, value)?,
+                local_name!("type") => self.type_ = NoiseType::parse(attr, value)?,
                 _ => (),
             }
         }
@@ -346,13 +346,12 @@ impl Filter for Turbulence {
         let mut affine = ctx.paffine();
         affine.invert();
 
-        let type_ = self.type_.get();
         let noise_generator = NoiseGenerator::new(
-            self.seed.get(),
-            self.base_frequency.get(),
-            self.num_octaves.get(),
-            type_,
-            self.stitch_tiles.get(),
+            self.seed,
+            self.base_frequency,
+            self.num_octaves,
+            self.type_,
+            self.stitch_tiles,
             f64::from(bounds.x1 - bounds.x0),
             f64::from(bounds.y1 - bounds.y0),
         );
@@ -380,7 +379,7 @@ impl Filter for Turbulence {
                             f64::from(y - bounds.y0),
                         );
 
-                        let v = match type_ {
+                        let v = match self.type_ {
                             NoiseType::FractalNoise => (v * 255.0 + 255.0) / 2.0,
                             NoiseType::Turbulence => v * 255.0,
                         };
@@ -401,7 +400,7 @@ impl Filter for Turbulence {
             }
         }
 
-        let cascaded = node.get_cascaded_values();
+        let cascaded = CascadedValues::new_from_node(node);
         let values = cascaded.get();
         // The generated color values are in the color space determined by
         // color-interpolation-filters.
@@ -413,7 +412,7 @@ impl Filter for Turbulence {
             };
 
         Ok(FilterResult {
-            name: self.base.result.borrow().clone(),
+            name: self.base.result.clone(),
             output: FilterOutput {
                 surface: SharedImageSurface::new(output_surface, surface_type)?,
                 bounds,
@@ -428,7 +427,7 @@ impl Filter for Turbulence {
 }
 
 impl StitchTiles {
-    fn parse(attr: Attribute, s: &str) -> Result<Self, NodeError> {
+    fn parse(attr: LocalName, s: &str) -> Result<Self, NodeError> {
         match s {
             "stitch" => Ok(StitchTiles::Stitch),
             "noStitch" => Ok(StitchTiles::NoStitch),
@@ -441,7 +440,7 @@ impl StitchTiles {
 }
 
 impl NoiseType {
-    fn parse(attr: Attribute, s: &str) -> Result<Self, NodeError> {
+    fn parse(attr: LocalName, s: &str) -> Result<Self, NodeError> {
         match s {
             "fractalNoise" => Ok(NoiseType::FractalNoise),
             "turbulence" => Ok(NoiseType::Turbulence),
