@@ -1,19 +1,20 @@
 use cairo::{self, MatrixTrait};
-use std::cell::Cell;
+use markup5ever::local_name;
 
-use crate::attributes::Attribute;
 use crate::bbox::BoundingBox;
 use crate::coord_units::CoordUnits;
 use crate::drawing_ctx::{CompositingAffines, DrawingCtx};
 use crate::error::RenderingError;
 use crate::length::{LengthHorizontal, LengthVertical};
-use crate::node::{NodeResult, NodeTrait, RsvgNode};
+use crate::node::{CascadedValues, NodeDraw, NodeResult, NodeTrait, RsvgNode};
 use crate::parsers::{Parse, ParseValue};
 use crate::property_bag::PropertyBag;
 use crate::property_defs::Opacity;
 use crate::rect::IRect;
 use crate::surface_utils::{
-    iterators::Pixels, shared_surface::SharedImageSurface, shared_surface::SurfaceType,
+    iterators::Pixels,
+    shared_surface::SharedImageSurface,
+    shared_surface::SurfaceType,
     ImageSurfaceDataExt,
 };
 use crate::unit_interval::UnitInterval;
@@ -22,30 +23,31 @@ coord_units!(MaskUnits, CoordUnits::ObjectBoundingBox);
 coord_units!(MaskContentUnits, CoordUnits::UserSpaceOnUse);
 
 pub struct NodeMask {
-    x: Cell<LengthHorizontal>,
-    y: Cell<LengthVertical>,
-    width: Cell<LengthHorizontal>,
-    height: Cell<LengthVertical>,
+    x: LengthHorizontal,
+    y: LengthVertical,
+    width: LengthHorizontal,
+    height: LengthVertical,
 
-    units: Cell<MaskUnits>,
-    content_units: Cell<MaskContentUnits>,
+    units: MaskUnits,
+    content_units: MaskContentUnits,
+}
+
+impl Default for NodeMask {
+    fn default() -> NodeMask {
+        NodeMask {
+            // these values are per the spec
+            x: LengthHorizontal::parse_str("-10%").unwrap(),
+            y: LengthVertical::parse_str("-10%").unwrap(),
+            width: LengthHorizontal::parse_str("120%").unwrap(),
+            height: LengthVertical::parse_str("120%").unwrap(),
+
+            units: MaskUnits::default(),
+            content_units: MaskContentUnits::default(),
+        }
+    }
 }
 
 impl NodeMask {
-    pub fn new() -> NodeMask {
-        NodeMask {
-            // these values are per the spec
-            x: Cell::new(LengthHorizontal::parse_str("-10%").unwrap()),
-            y: Cell::new(LengthVertical::parse_str("-10%").unwrap()),
-
-            width: Cell::new(LengthHorizontal::parse_str("120%").unwrap()),
-            height: Cell::new(LengthVertical::parse_str("120%").unwrap()),
-
-            units: Cell::new(MaskUnits::default()),
-            content_units: Cell::new(MaskContentUnits::default()),
-        }
-    }
-
     pub fn generate_cairo_mask(
         &self,
         mask_node: &RsvgNode,
@@ -61,13 +63,13 @@ impl NodeMask {
 
         let bbox_rect = bbox.rect.as_ref().unwrap();
 
-        let cascaded = mask_node.get_cascaded_values();
+        let cascaded = CascadedValues::new_from_node(mask_node);
         let values = cascaded.get();
 
         let mask_content_surface = draw_ctx.create_surface_for_toplevel_viewport()?;
 
-        let mask_units = CoordUnits::from(self.units.get());
-        let content_units = CoordUnits::from(self.content_units.get());
+        let mask_units = CoordUnits::from(self.units);
+        let content_units = CoordUnits::from(self.content_units);
 
         let (x, y, w, h) = {
             let params = if mask_units == CoordUnits::ObjectBoundingBox {
@@ -76,10 +78,10 @@ impl NodeMask {
                 draw_ctx.get_view_params()
             };
 
-            let x = self.x.get().normalize(&values, &params);
-            let y = self.y.get().normalize(&values, &params);
-            let w = self.width.get().normalize(&values, &params);
-            let h = self.height.get().normalize(&values, &params);
+            let x = self.x.normalize(&values, &params);
+            let y = self.y.normalize(&values, &params);
+            let w = self.width.normalize(&values, &params);
+            let h = self.height.normalize(&values, &params);
 
             (x, y, w, h)
         };
@@ -89,7 +91,7 @@ impl NodeMask {
         {
             let mask_cr = cairo::Context::new(&mask_content_surface);
             mask_cr.set_matrix(affines.for_temporary_surface);
-            mask_cr.transform(mask_node.get_transform());
+            mask_cr.transform(mask_node.borrow().get_transform());
 
             draw_ctx.push_cairo_context(mask_cr);
 
@@ -184,19 +186,21 @@ fn compute_luminance_to_alpha(
 }
 
 impl NodeTrait for NodeMask {
-    fn set_atts(&self, _: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::X => self.x.set(attr.parse(value)?),
-                Attribute::Y => self.y.set(attr.parse(value)?),
-                Attribute::Width => self
-                    .width
-                    .set(attr.parse_and_validate(value, LengthHorizontal::check_nonnegative)?),
-                Attribute::Height => self
-                    .height
-                    .set(attr.parse_and_validate(value, LengthVertical::check_nonnegative)?),
-                Attribute::MaskUnits => self.units.set(attr.parse(value)?),
-                Attribute::MaskContentUnits => self.content_units.set(attr.parse(value)?),
+                local_name!("x") => self.x = attr.parse(value)?,
+                local_name!("y") => self.y = attr.parse(value)?,
+                local_name!("width") => {
+                    self.width =
+                        attr.parse_and_validate(value, LengthHorizontal::check_nonnegative)?
+                }
+                local_name!("height") => {
+                    self.height =
+                        attr.parse_and_validate(value, LengthVertical::check_nonnegative)?
+                }
+                local_name!("maskUnits") => self.units = attr.parse(value)?,
+                local_name!("maskContentUnits") => self.content_units = attr.parse(value)?,
                 _ => (),
             }
         }
