@@ -1,78 +1,264 @@
 use cairo::{Matrix, MatrixTrait};
 use downcast_rs::*;
-use std::cell::{Cell, Ref, RefCell};
+use markup5ever::{local_name, LocalName};
+use std::cell::Ref;
 use std::collections::HashSet;
 use std::fmt;
-use std::rc::Rc;
 
-use crate::attributes::Attribute;
 use crate::cond::{RequiredExtensions, RequiredFeatures, SystemLanguage};
 use crate::css::CssRules;
 use crate::drawing_ctx::DrawingCtx;
 use crate::error::*;
+use crate::filters::Filter;
 use crate::parsers::Parse;
 use crate::properties::{ComputedValues, SpecifiedValue, SpecifiedValues};
 use crate::property_bag::PropertyBag;
 use crate::property_defs::Overflow;
-use crate::tree_utils::{self, NodeRef, NodeWeakRef};
 use locale_config::Locale;
+use rctree;
 
 /// Tree node with specific data
-pub type RsvgNode = NodeRef<NodeData>;
-pub type RsvgWeakNode = NodeWeakRef<NodeData>;
+pub type RsvgNode = rctree::Node<NodeData>;
+pub type RsvgWeakNode = rctree::WeakNode<NodeData>;
 
 /// Contents of a tree node
 pub struct NodeData {
     node_type: NodeType,
+    element_name: LocalName,
     id: Option<String>,    // id attribute from XML element
     class: Option<String>, // class attribute from XML element
-    specified_values: RefCell<SpecifiedValues>,
-    important_styles: RefCell<HashSet<Attribute>>,
-    result: RefCell<NodeResult>,
-    transform: Cell<Matrix>,
-    values: RefCell<ComputedValues>,
-    cond: Cell<bool>,
+    specified_values: SpecifiedValues,
+    important_styles: HashSet<LocalName>,
+    result: NodeResult,
+    transform: Matrix,
+    values: ComputedValues,
+    cond: bool,
+    style_attr: String,
     node_impl: Box<NodeTrait>,
-    style_attr: RefCell<String>,
-}
-
-impl NodeRef<NodeData> {
-    pub fn new(
-        node_type: NodeType,
-        parent: Option<&NodeRef<NodeData>>,
-        id: Option<&str>,
-        class: Option<&str>,
-        node_impl: Box<NodeTrait>,
-    ) -> NodeRef<NodeData> {
-        let data = NodeData {
-            node_type,
-            id: id.map(str::to_string),
-            class: class.map(str::to_string),
-            specified_values: RefCell::new(Default::default()),
-            important_styles: Default::default(),
-            transform: Cell::new(Matrix::identity()),
-            result: RefCell::new(Ok(())),
-            values: RefCell::new(ComputedValues::default()),
-            cond: Cell::new(true),
-            node_impl,
-            style_attr: RefCell::new(String::new()),
-        };
-
-        NodeRef(Rc::new(tree_utils::Node::new(data, parent)))
-    }
-
-    pub fn downgrade(&self) -> RsvgWeakNode {
-        Rc::downgrade(&self.0)
-    }
-
-    pub fn upgrade(weak: &RsvgWeakNode) -> Option<NodeRef<NodeData>> {
-        weak.upgrade().map(NodeRef)
-    }
 }
 
 impl NodeData {
-    pub fn get_impl<T: NodeTrait>(&self) -> Option<&T> {
-        (&self.node_impl).downcast_ref::<T>()
+    pub fn new(
+        node_type: NodeType,
+        element_name: LocalName,
+        id: Option<&str>,
+        class: Option<&str>,
+        node_impl: Box<NodeTrait>,
+    ) -> NodeData {
+        NodeData {
+            node_type,
+            element_name,
+            id: id.map(str::to_string),
+            class: class.map(str::to_string),
+            specified_values: Default::default(),
+            important_styles: Default::default(),
+            transform: Matrix::identity(),
+            result: Ok(()),
+            values: ComputedValues::default(),
+            cond: true,
+            style_attr: String::new(),
+            node_impl,
+        }
+    }
+
+    pub fn get_node_trait(&self) -> &NodeTrait {
+        self.node_impl.as_ref()
+    }
+
+    pub fn get_impl<T: NodeTrait>(&self) -> &T {
+        if let Some(t) = (&self.node_impl).downcast_ref::<T>() {
+            t
+        } else {
+            panic!("could not downcast");
+        }
+    }
+
+    pub fn get_type(&self) -> NodeType {
+        self.node_type
+    }
+
+    pub fn element_name(&self) -> &str {
+        self.element_name.as_ref()
+    }
+
+    pub fn get_id(&self) -> Option<&str> {
+        self.id.as_ref().map(String::as_str)
+    }
+
+    pub fn get_class(&self) -> Option<&str> {
+        self.class.as_ref().map(String::as_str)
+    }
+
+    pub fn get_cond(&self) -> bool {
+        self.cond
+    }
+
+    pub fn get_transform(&self) -> Matrix {
+        self.transform
+    }
+
+    pub fn is_overflow(&self) -> bool {
+        self.specified_values.is_overflow()
+    }
+
+    pub fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>, locale: &Locale) {
+        if self.node_impl.overflow_hidden() {
+            self.specified_values.overflow = SpecifiedValue::Specified(Overflow::Hidden);
+        }
+
+        self.save_style_attribute(pbag);
+
+        if let Err(e) = self
+            .set_transform_attribute(pbag)
+            .and_then(|_| self.set_conditional_processing_attributes(pbag, locale))
+            .and_then(|_| self.node_impl.set_atts(parent, pbag))
+            .and_then(|_| self.set_presentation_attributes(pbag))
+        {
+            self.set_error(e);
+        }
+
+        self.node_impl
+            .set_overridden_properties(&mut self.specified_values);
+    }
+
+    fn save_style_attribute(&mut self, pbag: &PropertyBag<'_>) {
+        for (attr, value) in pbag.iter() {
+            match attr {
+                local_name!("style") => self.style_attr.push_str(value),
+                _ => (),
+            }
+        }
+    }
+
+    fn set_transform_attribute(&mut self, pbag: &PropertyBag<'_>) -> Result<(), NodeError> {
+        for (attr, value) in pbag.iter() {
+            match attr {
+                local_name!("transform") => {
+                    return Matrix::parse_str(value).attribute(attr).and_then(|affine| {
+                        self.transform = affine;
+                        Ok(())
+                    });
+                }
+                _ => (),
+            }
+        }
+
+        Ok(())
+    }
+
+    fn set_conditional_processing_attributes(
+        &mut self,
+        pbag: &PropertyBag<'_>,
+        locale: &Locale,
+    ) -> Result<(), NodeError> {
+        let mut cond = self.cond;
+
+        for (attr, value) in pbag.iter() {
+            // FIXME: move this to "try {}" when we can bump the rustc version dependency
+            let mut parse = || {
+                match attr {
+                    local_name!("requiredExtensions") if cond => {
+                        cond = RequiredExtensions::from_attribute(value)
+                            .map(|RequiredExtensions(res)| res)?;
+                    }
+
+                    local_name!("requiredFeatures") if cond => {
+                        cond = RequiredFeatures::from_attribute(value)
+                            .map(|RequiredFeatures(res)| res)?;
+                    }
+
+                    local_name!("systemLanguage") if cond => {
+                        cond = SystemLanguage::from_attribute(value, locale)
+                            .map(|SystemLanguage(res)| res)?;
+                    }
+
+                    _ => {}
+                }
+
+                Ok(cond)
+            };
+
+            parse()
+                .map(|c| self.cond = c)
+                .map_err(|e| NodeError::attribute_error(attr, e))?;
+        }
+
+        Ok(())
+    }
+
+    /// Hands the pbag to the node's state, to apply the presentation attributes
+    fn set_presentation_attributes(&mut self, pbag: &PropertyBag<'_>) -> Result<(), NodeError> {
+        match self.specified_values.parse_presentation_attributes(pbag) {
+            Ok(_) => Ok(()),
+            Err(e) => {
+                // FIXME: we'll ignore errors here for now.
+                //
+                // If we set the node to be in error, we expose buggy handling of the
+                // enable-background property; we are not parsing it correctly. This
+                // causes tests/fixtures/reftests/bugs/587721-text-transform.svg to fail
+                // because it has enable-background="new 0 0 1179.75118 687.74173" in the
+                // toplevel svg element.
+                //
+                //   self.set_error(e);
+                //   return;
+
+                rsvg_log!("(attribute error: {})", e);
+                Ok(())
+            }
+        }
+    }
+
+    /// Applies the CSS rules that match into the node's specified_values
+    fn set_css_styles(&mut self, css_rules: &CssRules) {
+        for selector in &css_rules.get_matches(self) {
+            if let Some(decl_list) = css_rules.get_declarations(selector) {
+                for declaration in decl_list.iter() {
+                    self.specified_values
+                        .set_property_from_declaration(declaration, &mut self.important_styles);
+                }
+            }
+        }
+    }
+
+    /// Applies CSS styles from the saved value of the "style" attribute
+    fn set_style_attribute(&mut self) {
+        if !self.style_attr.is_empty() {
+            if let Err(e) = self
+                .specified_values
+                .parse_style_declarations(self.style_attr.as_str(), &mut self.important_styles)
+            {
+                self.set_error(e);
+            }
+
+            self.style_attr.clear();
+            self.style_attr.shrink_to_fit();
+        }
+    }
+
+    // Sets the node's specified values from the style-related attributes in the pbag.
+    // Also applies CSS rules in our limited way based on the node's tag/class/id.
+    pub fn set_style(&mut self, css_rules: &CssRules) {
+        self.set_css_styles(css_rules);
+        self.set_style_attribute();
+    }
+
+    fn set_error(&mut self, error: NodeError) {
+        self.result = Err(error);
+    }
+
+    pub fn is_in_error(&self) -> bool {
+        self.result.is_err()
+    }
+}
+
+impl fmt::Display for NodeData {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(
+            f,
+            "{:?} id={}",
+            self.get_type(),
+            self.get_id().unwrap_or("None")
+        )
     }
 }
 
@@ -84,14 +270,14 @@ impl NodeData {
 /// elsewhere in the SVG; it causes the instanced subtree to re-cascade from the computed values for
 /// the `<use>` element.
 ///
-/// This structure gets created by `Node.get_cascaded_values()`.  You can then call the `get()`
-/// method on the resulting `CascadedValues` to get a `&ComputedValues` whose fields you can access.
+/// You can then call the `get()` method on the resulting `CascadedValues` to get a
+/// `&ComputedValues` whose fields you can access.
 pub struct CascadedValues<'a> {
     inner: CascadedInner<'a>,
 }
 
 enum CascadedInner<'a> {
-    FromNode(Ref<'a, ComputedValues>),
+    FromNode(Ref<'a, NodeData>),
     FromValues(ComputedValues),
 }
 
@@ -104,7 +290,7 @@ impl<'a> CascadedValues<'a> {
     pub fn new(&self, node: &'a RsvgNode) -> CascadedValues<'a> {
         match self.inner {
             CascadedInner::FromNode(_) => CascadedValues {
-                inner: CascadedInner::FromNode(node.borrow().values.borrow()),
+                inner: CascadedInner::FromNode(node.borrow()),
             },
 
             CascadedInner::FromValues(ref v) => CascadedValues::new_from_values(node, v),
@@ -116,9 +302,9 @@ impl<'a> CascadedValues<'a> {
     /// This is to be used only in the toplevel drawing function, or in elements like `<marker>`
     /// that don't propagate their parent's cascade to their children.  All others should use
     /// `new()` to derive the cascade from an existing one.
-    fn new_from_node(node: &RsvgNode) -> CascadedValues<'_> {
+    pub fn new_from_node(node: &RsvgNode) -> CascadedValues<'_> {
         CascadedValues {
-            inner: CascadedInner::FromNode(node.borrow().values.borrow()),
+            inner: CascadedInner::FromNode(node.borrow()),
         }
     }
 
@@ -129,10 +315,7 @@ impl<'a> CascadedValues<'a> {
     /// `<use>`'s own cascade, not wih the element's original cascade.
     pub fn new_from_values(node: &'a RsvgNode, values: &ComputedValues) -> CascadedValues<'a> {
         let mut v = values.clone();
-        node.borrow()
-            .specified_values
-            .borrow()
-            .to_computed_values(&mut v);
+        node.borrow().specified_values.to_computed_values(&mut v);
 
         CascadedValues {
             inner: CascadedInner::FromValues(v),
@@ -145,7 +328,7 @@ impl<'a> CascadedValues<'a> {
     /// `ComputedValues` from the `CascadedValues` that got passed to `draw()`.
     pub fn get(&'a self) -> &'a ComputedValues {
         match self.inner {
-            CascadedInner::FromNode(ref r) => &*r,
+            CascadedInner::FromNode(ref node) => &node.values,
             CascadedInner::FromValues(ref v) => v,
         }
     }
@@ -156,12 +339,18 @@ pub trait NodeTrait: Downcast {
     /// Sets per-node attributes from the `pbag`
     ///
     /// Each node is supposed to iterate the `pbag`, and parse any attributes it needs.
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult;
+    fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult;
 
     /// Sets any special-cased properties that the node may have, that are different
     /// from defaults in the node's `SpecifiedValues`.
     fn set_overridden_properties(&self, _values: &mut SpecifiedValues) {}
 
+    /// Whether this node has overflow:hidden.
+    /// https://www.w3.org/TR/SVG/styling.html#UAStyleSheet
+    fn overflow_hidden(&self) -> bool {
+        false
+    }
+
     fn draw(
         &self,
         _node: &RsvgNode,
@@ -172,6 +361,11 @@ pub trait NodeTrait: Downcast {
         // by default nodes don't draw themselves
         Ok(())
     }
+
+    /// Returns the Filter trait if this node is a filter primitive
+    fn as_filter(&self) -> Option<&Filter> {
+        None
+    }
 }
 
 impl_downcast!(NodeTrait);
@@ -218,6 +412,7 @@ pub enum NodeType {
     Link,
     Marker,
     Mask,
+    NonRendering,
     Path,
     Pattern,
     PointLight,
@@ -236,302 +431,80 @@ pub enum NodeType {
     TSpan,
     Use,
 
-    // Filter primitives
-    FilterPrimitiveBlend,
-    FilterPrimitiveColorMatrix,
-    FilterPrimitiveComponentTransfer,
-    FilterPrimitiveComposite,
-    FilterPrimitiveConvolveMatrix,
-    FilterPrimitiveDiffuseLighting,
-    FilterPrimitiveDisplacementMap,
-    FilterPrimitiveFlood,
-    FilterPrimitiveGaussianBlur,
-    FilterPrimitiveImage,
-    FilterPrimitiveMerge,
-    FilterPrimitiveMergeNode,
-    FilterPrimitiveMorphology,
-    FilterPrimitiveOffset,
-    FilterPrimitiveSpecularLighting,
-    FilterPrimitiveTile,
-    FilterPrimitiveTurbulence,
+    // Filter primitives, these start with "Fe" as element names are e.g. "feBlend"
+    FeBlend,
+    FeColorMatrix,
+    FeComponentTransfer,
+    FeComposite,
+    FeConvolveMatrix,
+    FeDiffuseLighting,
+    FeDisplacementMap,
+    FeFlood,
+    FeGaussianBlur,
+    FeImage,
+    FeMerge,
+    FeMergeNode,
+    FeMorphology,
+    FeOffset,
+    FeSpecularLighting,
+    FeTile,
+    FeTurbulence,
 }
 
-impl NodeType {
-    pub fn element_name(&self) -> &'static str {
-        match self {
-            NodeType::Chars => "rsvg-chars", // Dummy element name for chars
-            NodeType::Circle => "circle",
-            NodeType::ClipPath => "clipPath",
-            NodeType::ComponentTransferFunctionA => "feFuncA",
-            NodeType::ComponentTransferFunctionB => "feFuncB",
-            NodeType::ComponentTransferFunctionG => "feFuncG",
-            NodeType::ComponentTransferFunctionR => "feFuncR",
-            NodeType::Defs => "defs",
-            NodeType::DistantLight => "feDistantLight",
-            NodeType::Ellipse => "ellipse",
-            NodeType::Filter => "filter",
-            NodeType::Group => "g",
-            NodeType::Image => "image",
-            NodeType::Line => "line",
-            NodeType::LinearGradient => "linearGradient",
-            NodeType::Link => "a",
-            NodeType::Marker => "marker",
-            NodeType::Mask => "mask",
-            NodeType::Path => "path",
-            NodeType::Pattern => "pattern",
-            NodeType::PointLight => "fePointight",
-            NodeType::Polygon => "polygon",
-            NodeType::Polyline => "polyline",
-            NodeType::RadialGradient => "radialGradient",
-            NodeType::Rect => "rect",
-            NodeType::SpotLight => "feSpotLight",
-            NodeType::Stop => "stop",
-            NodeType::Style => "style",
-            NodeType::Svg => "svg",
-            NodeType::Switch => "switch",
-            NodeType::Symbol => "symbol",
-            NodeType::Text => "text",
-            NodeType::TRef => "tref",
-            NodeType::TSpan => "tspan",
-            NodeType::Use => "use",
-            NodeType::FilterPrimitiveBlend => "feBlend",
-            NodeType::FilterPrimitiveColorMatrix => "feColorMatrix",
-            NodeType::FilterPrimitiveComponentTransfer => "feComponentTransfer",
-            NodeType::FilterPrimitiveComposite => "feComposite",
-            NodeType::FilterPrimitiveConvolveMatrix => "feConvolveMatrix",
-            NodeType::FilterPrimitiveDiffuseLighting => "feDiffuseLighting",
-            NodeType::FilterPrimitiveDisplacementMap => "feDisplacementMap",
-            NodeType::FilterPrimitiveFlood => "feFlood",
-            NodeType::FilterPrimitiveGaussianBlur => "feGaussianBlur",
-            NodeType::FilterPrimitiveImage => "feImage",
-            NodeType::FilterPrimitiveMerge => "feMerge",
-            NodeType::FilterPrimitiveMergeNode => "feMergeNode",
-            NodeType::FilterPrimitiveMorphology => "feMorphology",
-            NodeType::FilterPrimitiveOffset => "feOffset",
-            NodeType::FilterPrimitiveSpecularLighting => "feSpecularLighting",
-            NodeType::FilterPrimitiveTile => "feTile",
-            NodeType::FilterPrimitiveTurbulence => "feTurbulence",
-        }
-    }
+/// Helper trait for cascading recursively
+pub trait NodeCascade {
+    fn cascade(&mut self, values: &ComputedValues);
 }
 
-impl RsvgNode {
-    pub fn get_type(&self) -> NodeType {
-        self.borrow().node_type
-    }
-
-    pub fn get_id(&self) -> Option<&str> {
-        self.borrow().id.as_ref().map(String::as_str)
-    }
-
-    pub fn get_class(&self) -> Option<&str> {
-        self.borrow().class.as_ref().map(String::as_str)
-    }
-
-    pub fn get_transform(&self) -> Matrix {
-        self.borrow().transform.get()
-    }
-
-    pub fn get_cascaded_values(&self) -> CascadedValues<'_> {
-        CascadedValues::new_from_node(self)
-    }
-
-    pub fn cascade(&self, values: &ComputedValues) {
+impl NodeCascade for RsvgNode {
+    fn cascade(&mut self, values: &ComputedValues) {
         let mut values = values.clone();
-        self.borrow()
-            .specified_values
-            .borrow()
-            .to_computed_values(&mut values);
-        *self.borrow().values.borrow_mut() = values.clone();
 
-        for child in self.children() {
+        {
+            let mut node_mut = self.borrow_mut();
+
+            node_mut.specified_values.to_computed_values(&mut values);
+            node_mut.values = values.clone();
+        }
+
+        for mut child in self.children() {
             child.cascade(&values);
         }
     }
+}
 
-    pub fn set_styles_recursively(&self, node: &RsvgNode, css_rules: &CssRules) {
-        self.set_style(node, css_rules);
-
-        for child in self.children() {
-            child.set_styles_recursively(&child, css_rules);
-        }
-    }
-
-    pub fn get_cond(&self) -> bool {
-        self.borrow().cond.get()
-    }
-
-    fn set_transform_attribute(&self, pbag: &PropertyBag<'_>) -> Result<(), NodeError> {
-        for (attr, value) in pbag.iter() {
-            match attr {
-                Attribute::Transform => {
-                    return Matrix::parse_str(value)
-                        .attribute(Attribute::Transform)
-                        .and_then(|affine| Ok(self.borrow().transform.set(affine)));
-                }
-
-                _ => (),
-            }
-        }
-
-        Ok(())
-    }
-
-    fn save_style_attribute(&self, pbag: &PropertyBag<'_>) {
-        let mut style_attr = self.borrow().style_attr.borrow_mut();
-
-        for (attr, value) in pbag.iter() {
-            match attr {
-                Attribute::Style => style_attr.push_str(value),
-
-                _ => (),
-            }
-        }
-    }
-
-    pub fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>, locale: &Locale) {
-        self.save_style_attribute(pbag);
-
-        if let Err(e) = self
-            .set_transform_attribute(pbag)
-            .and_then(|_| self.parse_conditional_processing_attributes(pbag, locale))
-            .and_then(|_| self.borrow().node_impl.set_atts(node, pbag))
-            .and_then(|_| self.set_presentation_attributes(pbag))
-        {
-            self.set_error(e);
-        }
-    }
-
-    fn parse_conditional_processing_attributes(
+/// Helper trait for drawing recursively
+pub trait NodeDraw {
+    fn draw(
         &self,
-        pbag: &PropertyBag<'_>,
-        locale: &Locale,
-    ) -> Result<(), NodeError> {
-        let mut cond = self.get_cond();
+        cascaded: &CascadedValues<'_>,
+        draw_ctx: &mut DrawingCtx,
+        clipping: bool,
+    ) -> Result<(), RenderingError>;
 
-        for (attr, value) in pbag.iter() {
-            // FIXME: move this to "try {}" when we can bump the rustc version dependency
-            let mut parse = || {
-                match attr {
-                    Attribute::RequiredExtensions if cond => {
-                        cond = RequiredExtensions::from_attribute(value)
-                            .map(|RequiredExtensions(res)| res)?;
-                    }
-
-                    Attribute::RequiredFeatures if cond => {
-                        cond = RequiredFeatures::from_attribute(value)
-                            .map(|RequiredFeatures(res)| res)?;
-                    }
-
-                    Attribute::SystemLanguage if cond => {
-                        cond = SystemLanguage::from_attribute(value, locale)
-                            .map(|SystemLanguage(res)| res)?;
-                    }
-
-                    _ => {}
-                }
-
-                Ok(cond)
-            };
-
-            parse()
-                .map(|c| self.borrow().cond.set(c))
-                .map_err(|e| NodeError::attribute_error(attr, e))?;
-        }
-
-        Ok(())
-    }
-
-    /// Hands the pbag to the node's state, to apply the presentation attributes
-    fn set_presentation_attributes(&self, pbag: &PropertyBag<'_>) -> Result<(), NodeError> {
-        match self
-            .borrow()
-            .specified_values
-            .borrow_mut()
-            .parse_presentation_attributes(pbag)
-        {
-            Ok(_) => Ok(()),
-            Err(e) => {
-                // FIXME: we'll ignore errors here for now.
-                //
-                // If we set the node to be in error, we expose buggy handling of the
-                // enable-background property; we are not parsing it correctly. This
-                // causes tests/fixtures/reftests/bugs/587721-text-transform.svg to fail
-                // because it has enable-background="new 0 0 1179.75118 687.74173" in the
-                // toplevel svg element.
-                //
-                //   self.set_error(e);
-                //   return;
-
-                rsvg_log!("(attribute error: {})", e);
-                Ok(())
-            }
-        }
-    }
-
-    /// Applies the CSS rules that match into the node's specified_values
-    fn set_css_styles(&self, node: &RsvgNode, css_rules: &CssRules) {
-        let mut specified_values = self.borrow().specified_values.borrow_mut();
-        let mut important_styles = self.borrow().important_styles.borrow_mut();
-
-        for selector in &css_rules.get_matches(node) {
-            if let Some(decl_list) = css_rules.get_declarations(selector) {
-                for declaration in decl_list.iter() {
-                    specified_values
-                        .set_property_from_declaration(declaration, &mut important_styles);
-                }
-            }
-        }
-    }
-
-    /// Applies CSS styles from the saved value of the "style" attribute
-    fn set_style_attribute(&self) {
-        let mut style_attr = self.borrow().style_attr.borrow_mut();
-
-        if !style_attr.is_empty() {
-            let mut important_styles = self.borrow().important_styles.borrow_mut();
-
-            if let Err(e) = self
-                .borrow()
-                .specified_values
-                .borrow_mut()
-                .parse_style_declarations(style_attr.as_str(), &mut important_styles)
-            {
-                self.set_error(e);
-            }
-
-            style_attr.clear();
-            style_attr.shrink_to_fit();
-        }
-    }
-
-    // Sets the node's specified values from the style-related attributes in the pbag.
-    // Also applies CSS rules in our limited way based on the node's tag/class/id.
-    fn set_style(&self, node: &RsvgNode, css_rules: &CssRules) {
-        self.set_css_styles(node, css_rules);
-        self.set_style_attribute();
-    }
-
-    pub fn set_overridden_properties(&self) {
-        let mut specified_values = self.borrow().specified_values.borrow_mut();
-        self.borrow()
-            .node_impl
-            .set_overridden_properties(&mut specified_values);
-    }
-
-    pub fn draw(
+    fn draw_children(
+        &self,
+        cascaded: &CascadedValues<'_>,
+        draw_ctx: &mut DrawingCtx,
+        clipping: bool,
+    ) -> Result<(), RenderingError>;
+}
+
+impl NodeDraw for RsvgNode {
+    fn draw(
         &self,
-        node: &RsvgNode,
         cascaded: &CascadedValues<'_>,
         draw_ctx: &mut DrawingCtx,
         clipping: bool,
     ) -> Result<(), RenderingError> {
-        if !self.is_in_error() {
+        if !self.borrow().is_in_error() {
             draw_ctx.with_saved_matrix(&mut |dc| {
                 let cr = dc.get_cairo_context();
-                cr.transform(self.get_transform());
+                cr.transform(self.borrow().get_transform());
 
-                self.borrow().node_impl.draw(node, cascaded, dc, clipping)
+                self.borrow()
+                    .get_node_trait()
+                    .draw(self, cascaded, dc, clipping)
             })
         } else {
             rsvg_log!("(not rendering element {} because it is in error)", self);
@@ -540,33 +513,7 @@ impl RsvgNode {
         }
     }
 
-    pub fn set_error(&self, error: NodeError) {
-        rsvg_log!("(attribute error for {}:\n  {})", self, error);
-
-        *self.borrow().result.borrow_mut() = Err(error);
-    }
-
-    pub fn is_in_error(&self) -> bool {
-        self.borrow().result.borrow().is_err()
-    }
-
-    pub fn with_impl<T, F, U>(&self, f: F) -> U
-    where
-        T: NodeTrait,
-        F: FnOnce(&T) -> U,
-    {
-        if let Some(t) = (&self.borrow().node_impl).downcast_ref::<T>() {
-            f(t)
-        } else {
-            panic!("could not downcast");
-        }
-    }
-
-    pub fn get_impl<T: NodeTrait>(&self) -> Option<&T> {
-        self.borrow().get_impl()
-    }
-
-    pub fn draw_children(
+    fn draw_children(
         &self,
         cascaded: &CascadedValues<'_>,
         draw_ctx: &mut DrawingCtx,
@@ -582,24 +529,4 @@ impl RsvgNode {
 
         Ok(())
     }
-
-    pub fn is_overflow(&self) -> bool {
-        self.borrow().specified_values.borrow().is_overflow()
-    }
-
-    pub fn set_overflow_hidden(&self) {
-        let mut specified_values = self.borrow().specified_values.borrow_mut();
-        specified_values.overflow = SpecifiedValue::Specified(Overflow::Hidden);
-    }
-}
-
-impl fmt::Display for RsvgNode {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(
-            f,
-            "{:?} id={}",
-            self.get_type(),
-            self.get_id().unwrap_or("None")
-        )
-    }
 }
