@@ -1,11 +1,11 @@
 use cairo;
 use cairo::{MatrixTrait, PatternTrait};
+use markup5ever::local_name;
 use std::cell::RefCell;
 use std::f64;
 
 use crate::allowed_url::Fragment;
 use crate::aspect_ratio::*;
-use crate::attributes::Attribute;
 use crate::bbox::*;
 use crate::coord_units::CoordUnits;
 use crate::drawing_ctx::{DrawingCtx, NodeStack};
@@ -13,7 +13,7 @@ use crate::error::{AttributeResultExt, RenderingError};
 use crate::float_eq_cairo::ApproxEqCairo;
 use crate::length::*;
 use crate::node::*;
-use crate::paint_server::PaintSource;
+use crate::paint_server::{PaintSource, Resolve};
 use crate::parsers::ParseValue;
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
@@ -24,8 +24,14 @@ use crate::viewbox::*;
 coord_units!(PatternUnits, CoordUnits::ObjectBoundingBox);
 coord_units!(PatternContentUnits, CoordUnits::UserSpaceOnUse);
 
-#[derive(Clone)]
-pub struct Pattern {
+macro_rules! fallback_to (
+    ($dest:expr, $default:expr) => (
+        $dest = $dest.take ().or ($default)
+    );
+);
+
+#[derive(Clone, Default)]
+pub struct NodePattern {
     pub units: Option<PatternUnits>,
     pub content_units: Option<PatternContentUnits>,
     // This Option<Option<ViewBox>> is a bit strange.  We want a field
@@ -44,69 +50,46 @@ pub struct Pattern {
     // Point back to our corresponding node, or to the fallback node which has children.
     // If the value is None, it means we are fully resolved and didn't find any children
     // among the fallbacks.
-    pub node: Option<RsvgWeakNode>,
+    pub node: RefCell<Option<RsvgWeakNode>>,
 }
 
-impl Default for Pattern {
-    fn default() -> Pattern {
-        // These are per the spec
-
-        Pattern {
-            units: Some(PatternUnits::default()),
-            content_units: Some(PatternContentUnits::default()),
-            vbox: Some(None),
-            preserve_aspect_ratio: Some(AspectRatio::default()),
-            affine: Some(cairo::Matrix::identity()),
-            fallback: None,
-            x: Some(Default::default()),
-            y: Some(Default::default()),
-            width: Some(Default::default()),
-            height: Some(Default::default()),
-            node: None,
+impl NodeTrait for NodePattern {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        for (attr, value) in pbag.iter() {
+            match attr {
+                local_name!("patternUnits") => self.units = Some(attr.parse(value)?),
+                local_name!("patternContentUnits") => self.content_units = Some(attr.parse(value)?),
+                local_name!("viewBox") => self.vbox = Some(Some(attr.parse(value)?)),
+                local_name!("preserveAspectRatio") => {
+                    self.preserve_aspect_ratio = Some(attr.parse(value)?)
+                }
+                local_name!("patternTransform") => self.affine = Some(attr.parse(value)?),
+                local_name!("xlink:href") => {
+                    self.fallback = Some(Fragment::parse(value).attribute(attr)?);
+                }
+                local_name!("x") => self.x = Some(attr.parse(value)?),
+                local_name!("y") => self.y = Some(attr.parse(value)?),
+                local_name!("width") => {
+                    self.width =
+                        Some(attr.parse_and_validate(value, LengthHorizontal::check_nonnegative)?)
+                }
+                local_name!("height") => {
+                    self.height =
+                        Some(attr.parse_and_validate(value, LengthVertical::check_nonnegative)?)
+                }
+                _ => (),
+            }
         }
+
+        Ok(())
+    }
+
+    fn overflow_hidden(&self) -> bool {
+        true
     }
 }
 
-// All of the Pattern's fields are Option<foo> values, because
-// those fields can be omitted in the SVG file.  We need to resolve
-// them to default values, or to fallback values that come from
-// another Pattern.
-//
-// For the fallback case, this would need something like
-//
-//    if self.foo.is_none () { self.foo = fallback.foo; }
-//
-// And for the default case, it would be like
-//    if self.foo.is_none () { self.foo = Some (default_value); }
-//
-// Both can be replaced by
-//
-//    self.foo = self.foo.take ().or (bar);
-//
-// So we define a macro for that.
-macro_rules! fallback_to (
-    ($dest:expr, $default:expr) => (
-        $dest = $dest.take ().or ($default)
-    );
-);
-
-impl Pattern {
-    fn unresolved() -> Pattern {
-        Pattern {
-            units: None,
-            content_units: None,
-            vbox: None,
-            preserve_aspect_ratio: None,
-            affine: None,
-            fallback: None,
-            x: None,
-            y: None,
-            width: None,
-            height: None,
-            node: None,
-        }
-    }
-
+impl Resolve for NodePattern {
     fn is_resolved(&self) -> bool {
         self.units.is_some()
             && self.content_units.is_some()
@@ -120,22 +103,7 @@ impl Pattern {
             && self.children_are_resolved()
     }
 
-    fn children_are_resolved(&self) -> bool {
-        if let Some(ref weak) = self.node {
-            let strong_node = &weak.clone().upgrade().unwrap();
-            strong_node.has_children()
-        } else {
-            // We are an empty pattern; there is nothing further that
-            // can be resolved for children.
-            true
-        }
-    }
-
-    fn resolve_from_defaults(&mut self) {
-        self.resolve_from_fallback(&Pattern::default());
-    }
-
-    fn resolve_from_fallback(&mut self, fallback: &Pattern) {
+    fn resolve_from_fallback(&mut self, fallback: &NodePattern) {
         fallback_to!(self.units, fallback.units);
         fallback_to!(self.content_units, fallback.content_units);
         fallback_to!(self.vbox, fallback.vbox);
@@ -149,79 +117,29 @@ impl Pattern {
         self.fallback = fallback.fallback.clone();
 
         if !self.children_are_resolved() {
-            if fallback.node.is_some() {
-                self.node = fallback.node.clone();
+            if fallback.node.borrow().is_some() {
+                *self.node.borrow_mut() = fallback.node.borrow().clone();
             } else {
-                self.node = None;
+                *self.node.borrow_mut() = None;
             }
         }
     }
-}
 
-pub struct NodePattern {
-    pattern: RefCell<Pattern>,
-}
-
-impl NodePattern {
-    pub fn new() -> NodePattern {
-        NodePattern {
-            pattern: RefCell::new(Pattern::unresolved()),
-        }
-    }
-}
-
-impl NodeTrait for NodePattern {
-    fn set_atts(&self, node: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
-        // pattern element has overflow:hidden
-        // https://www.w3.org/TR/SVG/styling.html#UAStyleSheet
-        node.set_overflow_hidden();
-
-        let mut p = self.pattern.borrow_mut();
-
-        p.node = Some(node.downgrade());
-
-        for (attr, value) in pbag.iter() {
-            match attr {
-                Attribute::PatternUnits => p.units = Some(attr.parse(value)?),
-
-                Attribute::PatternContentUnits => p.content_units = Some(attr.parse(value)?),
-
-                Attribute::ViewBox => p.vbox = Some(Some(attr.parse(value)?)),
-
-                Attribute::PreserveAspectRatio => {
-                    p.preserve_aspect_ratio = Some(attr.parse(value)?)
-                }
-
-                Attribute::PatternTransform => p.affine = Some(attr.parse(value)?),
-
-                Attribute::XlinkHref => {
-                    p.fallback = Some(Fragment::parse(value).attribute(Attribute::XlinkHref)?);
-                }
-
-                Attribute::X => p.x = Some(attr.parse(value)?),
-
-                Attribute::Y => p.y = Some(attr.parse(value)?),
-
-                Attribute::Width => {
-                    p.width =
-                        Some(attr.parse_and_validate(value, LengthHorizontal::check_nonnegative)?)
-                }
-
-                Attribute::Height => {
-                    p.height =
-                        Some(attr.parse_and_validate(value, LengthVertical::check_nonnegative)?)
-                }
-
-                _ => (),
-            }
-        }
-
-        Ok(())
+    fn resolve_from_defaults(&mut self) {
+        fallback_to!(self.units, Some(PatternUnits::default()));
+        fallback_to!(self.content_units, Some(PatternContentUnits::default()));
+        fallback_to!(self.vbox, Some(None));
+        fallback_to!(self.preserve_aspect_ratio, Some(AspectRatio::default()));
+        fallback_to!(self.affine, Some(cairo::Matrix::identity()));
+        fallback_to!(self.x, Some(Default::default()));
+        fallback_to!(self.y, Some(Default::default()));
+        fallback_to!(self.width, Some(Default::default()));
+        fallback_to!(self.height, Some(Default::default()));
     }
 }
 
 impl PaintSource for NodePattern {
-    type Source = Pattern;
+    type Source = NodePattern;
 
     fn resolve(
         &self,
@@ -229,9 +147,9 @@ impl PaintSource for NodePattern {
         draw_ctx: &mut DrawingCtx,
         _bbox: &BoundingBox,
     ) -> Result<Option<Self::Source>, RenderingError> {
-        let node_pattern = node.get_impl::<NodePattern>().unwrap();
-        let pattern = &*node_pattern.pattern.borrow();
-        let mut result = pattern.clone();
+        *self.node.borrow_mut() = Some(node.downgrade());
+
+        let mut result = node.borrow().get_impl::<NodePattern>().clone();
         let mut stack = NodeStack::new();
 
         while !result.is_resolved() {
@@ -246,9 +164,8 @@ impl PaintSource for NodePattern {
                     return Err(RenderingError::CircularReference);
                 }
 
-                a_node.with_impl(|i: &NodePattern| {
-                    result.resolve_from_fallback(&*i.pattern.borrow())
-                });
+                let node_data = a_node.borrow();
+                result.resolve_from_fallback(&node_data.get_impl::<NodePattern>());
 
                 stack.push(a_node);
             } else {
@@ -269,7 +186,7 @@ impl PaintSource for NodePattern {
     ) -> Result<bool, RenderingError> {
         assert!(pattern.is_resolved());
 
-        if pattern.node.is_none() {
+        if pattern.node.borrow().is_none() {
             // This means we didn't find any children among the fallbacks,
             // so there is nothing to render.
             return Ok(false);
@@ -418,8 +335,8 @@ impl PaintSource for NodePattern {
         // Set up transformations to be determined by the contents units
 
         // Draw everything
-        let pattern_node = RsvgNode::upgrade(pattern.node.as_ref().unwrap()).unwrap();
-        let pattern_cascaded = pattern_node.get_cascaded_values();
+        let pattern_node = pattern.node.borrow().as_ref().unwrap().upgrade().unwrap();
+        let pattern_cascaded = CascadedValues::new_from_node(&pattern_node);
         let pattern_values = pattern_cascaded.get();
 
         cr_pattern.set_matrix(caffine);
@@ -449,13 +366,26 @@ impl PaintSource for NodePattern {
     }
 }
 
+impl NodePattern {
+    fn children_are_resolved(&self) -> bool {
+        if let Some(ref weak) = *self.node.borrow() {
+            let strong_node = &weak.clone().upgrade().unwrap();
+            strong_node.has_children()
+        } else {
+            // We are an empty pattern; there is nothing further that
+            // can be resolved for children.
+            true
+        }
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
 
     #[test]
     fn pattern_resolved_from_defaults_is_really_resolved() {
-        let mut pat = Pattern::unresolved();
+        let mut pat = NodePattern::default();
 
         pat.resolve_from_defaults();
         assert!(pat.is_resolved());
