@@ -1,11 +1,11 @@
 use glib::translate::*;
+use markup5ever::local_name;
 use pango::{self, ContextExt, FontMapExt, LayoutExt};
 use pango_sys;
 use pangocairo;
-use std::cell::{Cell, RefCell};
+use std::cell::RefCell;
 
 use crate::allowed_url::Fragment;
-use crate::attributes::Attribute;
 use crate::bbox::BoundingBox;
 use crate::drawing_ctx::DrawingCtx;
 use crate::error::{AttributeResultExt, RenderingError};
@@ -15,6 +15,7 @@ use crate::length::*;
 use crate::node::{CascadedValues, NodeResult, NodeTrait, NodeType, RsvgNode};
 use crate::parsers::ParseValue;
 use crate::properties::ComputedValues;
+use crate::property_bag::PropertyBag;
 use crate::property_defs::{
     Direction,
     FontStretch,
@@ -27,7 +28,6 @@ use crate::property_defs::{
     XmlLang,
     XmlSpace,
 };
-use crate::property_bag::PropertyBag;
 use crate::space::{xml_space_normalize, NormalizeDefault, XmlSpaceNormalize};
 
 /// An absolutely-positioned array of `Span`s
@@ -419,21 +419,36 @@ fn children_to_chunks(
     depth: usize,
 ) {
     for child in node.children() {
-        match child.get_type() {
-            NodeType::Chars => child.with_impl(|chars: &NodeChars| {
+        match child.borrow().get_type() {
+            NodeType::Chars => {
                 let values = cascaded.get();
-                chars.to_chunks(&child, values, chunks, dx, dy, depth);
-            }),
+                child
+                    .borrow()
+                    .get_impl::<NodeChars>()
+                    .to_chunks(&child, values, chunks, dx, dy, depth);
+            }
 
-            NodeType::TSpan => child.with_impl(|tspan: &NodeTSpan| {
+            NodeType::TSpan => {
                 let cascaded = CascadedValues::new(cascaded, &child);
-                tspan.to_chunks(&child, &cascaded, draw_ctx, chunks, depth + 1);
-            }),
+                child.borrow().get_impl::<NodeTSpan>().to_chunks(
+                    &child,
+                    &cascaded,
+                    draw_ctx,
+                    chunks,
+                    depth + 1,
+                );
+            }
 
-            NodeType::TRef => child.with_impl(|tref: &NodeTRef| {
+            NodeType::TRef => {
                 let cascaded = CascadedValues::new(cascaded, &child);
-                tref.to_chunks(&child, &cascaded, draw_ctx, chunks, depth + 1);
-            }),
+                child.borrow().get_impl::<NodeTRef>().to_chunks(
+                    &child,
+                    &cascaded,
+                    draw_ctx,
+                    chunks,
+                    depth + 1,
+                );
+            }
 
             _ => (),
         }
@@ -540,28 +555,20 @@ impl NodeChars {
 }
 
 impl NodeTrait for NodeChars {
-    fn set_atts(&self, _: &RsvgNode, _: &PropertyBag<'_>) -> NodeResult {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, _: &PropertyBag<'_>) -> NodeResult {
         Ok(())
     }
 }
 
+#[derive(Default)]
 pub struct NodeText {
-    x: Cell<LengthHorizontal>,
-    y: Cell<LengthVertical>,
-    dx: Cell<Option<LengthHorizontal>>,
-    dy: Cell<Option<LengthVertical>>,
+    x: LengthHorizontal,
+    y: LengthVertical,
+    dx: Option<LengthHorizontal>,
+    dy: Option<LengthVertical>,
 }
 
 impl NodeText {
-    pub fn new() -> NodeText {
-        NodeText {
-            x: Cell::new(Default::default()),
-            y: Cell::new(Default::default()),
-            dx: Cell::new(None),
-            dy: Cell::new(None),
-        }
-    }
-
     fn make_chunks(
         &self,
         node: &RsvgNode,
@@ -569,27 +576,20 @@ impl NodeText {
         draw_ctx: &mut DrawingCtx,
     ) -> Vec<Chunk> {
         let mut chunks = Vec::new();
-
-        let x = self.x.get();
-        let y = self.y.get();
-        let dx = self.dx.get();
-        let dy = self.dy.get();
-
-        chunks.push(Chunk::new(cascaded.get(), Some(x), Some(y)));
-
-        children_to_chunks(&mut chunks, node, cascaded, draw_ctx, dx, dy, 0);
+        chunks.push(Chunk::new(cascaded.get(), Some(self.x), Some(self.y)));
+        children_to_chunks(&mut chunks, node, cascaded, draw_ctx, self.dx, self.dy, 0);
         chunks
     }
 }
 
 impl NodeTrait for NodeText {
-    fn set_atts(&self, _: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::X => self.x.set(attr.parse(value)?),
-                Attribute::Y => self.y.set(attr.parse(value)?),
-                Attribute::Dx => self.dx.set(attr.parse(value).map(Some)?),
-                Attribute::Dy => self.dy.set(attr.parse(value).map(Some)?),
+                local_name!("x") => self.x = attr.parse(value)?,
+                local_name!("y") => self.y = attr.parse(value)?,
+                local_name!("dx") => self.dx = attr.parse(value).map(Some)?,
+                local_name!("dy") => self.dy = attr.parse(value).map(Some)?,
                 _ => (),
             }
         }
@@ -607,8 +607,8 @@ impl NodeTrait for NodeText {
         let values = cascaded.get();
         let params = draw_ctx.get_view_params();
 
-        let mut x = self.x.get().normalize(values, &params);
-        let mut y = self.y.get().normalize(values, &params);
+        let mut x = self.x.normalize(values, &params);
+        let mut y = self.y.normalize(values, &params);
 
         let chunks = self.make_chunks(node, cascaded, draw_ctx);
 
@@ -646,17 +646,12 @@ impl NodeTrait for NodeText {
     }
 }
 
+#[derive(Default)]
 pub struct NodeTRef {
-    link: RefCell<Option<Fragment>>,
+    link: Option<Fragment>,
 }
 
 impl NodeTRef {
-    pub fn new() -> NodeTRef {
-        NodeTRef {
-            link: RefCell::new(Default::default()),
-        }
-    }
-
     fn to_chunks(
         &self,
         node: &RsvgNode,
@@ -665,14 +660,11 @@ impl NodeTRef {
         chunks: &mut Vec<Chunk>,
         depth: usize,
     ) {
-        let link = self.link.borrow();
-
-        if link.is_none() {
+        if self.link.is_none() {
             return;
         }
 
-        let link = link.as_ref().unwrap();
-
+        let link = self.link.as_ref().unwrap();
         let values = cascaded.get();
 
         if let Some(acquired) = draw_ctx.acquired_nodes().get_node(link) {
@@ -695,23 +687,22 @@ fn extract_chars_children_to_chunks_recursively(
     depth: usize,
 ) {
     for child in node.children() {
-        match child.get_type() {
-            NodeType::Chars => child.with_impl(|chars: &NodeChars| {
-                chars.to_chunks(&child, values, chunks, None, None, depth);
-            }),
-
+        match child.borrow().get_type() {
+            NodeType::Chars => child
+                .borrow()
+                .get_impl::<NodeChars>()
+                .to_chunks(&child, values, chunks, None, None, depth),
             _ => extract_chars_children_to_chunks_recursively(chunks, &child, values, depth + 1),
         }
     }
 }
 
 impl NodeTrait for NodeTRef {
-    fn set_atts(&self, _: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::XlinkHref => {
-                    *self.link.borrow_mut() =
-                        Some(Fragment::parse(value).attribute(Attribute::XlinkHref)?)
+                local_name!("xlink:href") => {
+                    self.link = Some(Fragment::parse(value).attribute(attr)?)
                 }
                 _ => (),
             }
@@ -721,23 +712,15 @@ impl NodeTrait for NodeTRef {
     }
 }
 
+#[derive(Default)]
 pub struct NodeTSpan {
-    x: Cell<Option<LengthHorizontal>>,
-    y: Cell<Option<LengthVertical>>,
-    dx: Cell<Option<LengthHorizontal>>,
-    dy: Cell<Option<LengthVertical>>,
+    x: Option<LengthHorizontal>,
+    y: Option<LengthVertical>,
+    dx: Option<LengthHorizontal>,
+    dy: Option<LengthVertical>,
 }
 
 impl NodeTSpan {
-    pub fn new() -> NodeTSpan {
-        NodeTSpan {
-            x: Cell::new(Default::default()),
-            y: Cell::new(Default::default()),
-            dx: Cell::new(None),
-            dy: Cell::new(None),
-        }
-    }
-
     fn to_chunks(
         &self,
         node: &RsvgNode,
@@ -746,29 +729,24 @@ impl NodeTSpan {
         chunks: &mut Vec<Chunk>,
         depth: usize,
     ) {
-        let x = self.x.get();
-        let y = self.y.get();
-        let dx = self.dx.get();
-        let dy = self.dy.get();
-
-        if x.is_some() || y.is_some() {
+        if self.x.is_some() || self.y.is_some() {
             // Any absolute position creates a new chunk
             let values = cascaded.get();
-            chunks.push(Chunk::new(values, x, y));
+            chunks.push(Chunk::new(values, self.x, self.y));
         }
 
-        children_to_chunks(chunks, node, cascaded, draw_ctx, dx, dy, depth);
+        children_to_chunks(chunks, node, cascaded, draw_ctx, self.dx, self.dy, depth);
     }
 }
 
 impl NodeTrait for NodeTSpan {
-    fn set_atts(&self, _: &RsvgNode, pbag: &PropertyBag<'_>) -> NodeResult {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr {
-                Attribute::X => self.x.set(attr.parse(value).map(Some)?),
-                Attribute::Y => self.y.set(attr.parse(value).map(Some)?),
-                Attribute::Dx => self.dx.set(attr.parse(value).map(Some)?),
-                Attribute::Dy => self.dy.set(attr.parse(value).map(Some)?),
+                local_name!("x") => self.x = attr.parse(value).map(Some)?,
+                local_name!("y") => self.y = attr.parse(value).map(Some)?,
+                local_name!("dx") => self.dx = attr.parse(value).map(Some)?,
+                local_name!("dy") => self.dy = attr.parse(value).map(Some)?,
                 _ => (),
             }
         }
